from textual.app import App, ComposeResult
from textual.containers import Container
from textual.widgets import Footer, Label
from textual.widgets._header import HeaderTitle
from textual.reactive import reactive

from pypod import config
from pypod.ui import (
    ProgressDisplay,
    PlaylistListView,
    Controls,
    Header,
)
from pypod.player import Pod


class PyPodApp(App):
    """A Python terminal music player app."""

    CSS_PATH = "style.css"
    BINDINGS = [
        ("q", "quit", "quit"),
        ("p", "toggle_play", "play/pause"),
        ("n", "play_next", "next"),
        ("b", "play_prev", "back (previous)"),
    ]

    song_title = reactive("♫ not playing ♫")

    def __init__(self, player: Pod, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.player = player

    def on_mount(self):
        # ♫
        self.query_one(HeaderTitle).text = "🎵 PyPod"
        self.set_interval(0.5, self.change_song)

    def watch_song_title(self, title: str):
        if not title.startswith("♫"):
            title = f"♫ {title} ..."
        self.query_one("#title").update(title)
        song_index = self.player.playlist.position
        self.query_one("#playlist").set_index(song_index)

    def change_song(self):
        # TODO: check the end of playlist
        if self.player.auto_finish.is_set():
            self.action_play_next()

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        table = PlaylistListView(
            id="playlist",
            playlist=self.player.playlist,
        )
        yield Container(
            Label(id="title"),
            ProgressDisplay(id="prog"),
            Controls(),
            table,
        )

        yield Footer()

    def action_toggle_play(self):
        progress: ProgressDisplay = self.query_one("#prog")
        if self.player.is_playing:
            progress.pause()
            self.player.pause()
            self.query_one("#play").label = "▶"
        else:
            self.player.play()
            self.query_one("#play").label = "▮▮"
            if progress.timer is None:
                progress.display_progress(self.player.song)
            else:
                progress.resume()
        self.song_title = self.player.song.name

    def action_play_next(self):
        progress: ProgressDisplay = self.query_one("#prog")
        self.player.next()
        self.query_one("#play").label = "▮▮"
        self.song_title = self.player.song.name
        progress.display_progress(self.player.song)

    def action_play_prev(self):
        progress: ProgressDisplay = self.query_one("#prog")
        self.player.prev()
        self.query_one("#play").label = "▮▮"
        self.song_title = self.player.song.name
        progress.display_progress(self.player.song)

    async def action_quit(self):
        """Quit the app with necessary cleanup."""
        self.player.exit()
        self.exit()


def create_app_debug():
    player = Pod()
    playlist = player.generate_playlist(config.ASSETS_DIR)
    player.load(playlist)
    app = PyPodApp(player=player)
    return app


app = create_app_debug()


if __name__ == "__main__":
    app.run()



from rich.progress import Progress, Text, Task, TextColumn, BarColumn
from textual.app import ComposeResult
from textual.widgets._header import HeaderTitle
from textual.widgets import ListView, ListItem, Static, Label, DataTable
from textual.widgets import Header as Header_
from textual.widgets import Button, Static

from pypod.song import Song
from pypod.utils import sec_to_time


class Controls(Static):
    """A buttons widget to control playback."""

    def compose(self) -> ComposeResult:
        """Create child widgets of a stopwatch."""
        yield Button("❮❮", id="prev", variant="success")
        yield Button("▶", id="play", variant="warning")
        yield Button("❯❯", id="next", variant="success")

    def on_button_pressed(self, event: Button.Pressed):
        """Event handler called when a button is pressed."""
        button = event.button
        if button.id == "play":
            self.app.action_toggle_play()
        elif button.id == "prev":
            self.app.action_play_prev()
        elif button.id == "next":
            self.app.action_play_next()


class Header(Header_):
    DEFAULT_CSS = """
    Header {
        dock: top;
        width: 100%;
        background: $secondary-background;
        color: $text;
        height: 1;
    }
    Header.-tall {
        height: 1;
    }
    """

    def compose(self):
        yield HeaderTitle()

    def on_click(self):
        pass


class PlaylistListView(Static):
    def __init__(self, playlist, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.playlist = playlist
        self._list = ListView()

    def on_mount(self):
        # TODO: add header somehow
        for i, s in enumerate(self.playlist, start=1):
            duration = sec_to_time(s.duration)
            self._list.append(
                ListItem(
                    Label(f"{i}"),
                    Label(f"{s}", classes="song-name"),
                    Label(f"{duration}", classes="right"),
                )
            )

    def set_index(self, highlight: int = 0):
        self._list.index = highlight

    def compose(self):
        yield self._list


class PlaylistTable(Static):
    def __init__(self, playlist, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.playlist = playlist
        self.table = DataTable()

    def on_mount(self):
        # table = self.query_one(DataTable)
        table = self.table
        table.add_columns("#", "Name", "Duration")
        for i, s in enumerate(self.playlist, start=1):
            duration = sec_to_time(s.duration)
            table.add_row(f"{i}", f"{s}", f"{duration}")

    def compose(self):
        yield self.table


class ElapsedColumn(TextColumn):
    def __init__(self, *args, **kwargs):
        return super().__init__("", *args, **kwargs)

    def render(self, task: Task) -> Text:
        if task.description == "dummy":
            return "--:--"
        return sec_to_time(task.completed)


class LeftColumn(TextColumn):
    pass


class DurationColumn(TextColumn):
    def __init__(self, *args, **kwargs):
        return super().__init__("", *args, **kwargs)

    def render(self, task: Task) -> Text:
        if task.description == "dummy":
            return "--:--"
        return sec_to_time(task.total)


class ProgressDisplay(Static):
    """A widget to display song's progress."""

    INTERVAL = 0.1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.start_col = ElapsedColumn()
        self.end_col = DurationColumn()
        self.bar = BarColumn()
        self.p = Progress(
            self.start_col,
            self.bar,
            self.end_col,
        )
        self.task_id = self.p.add_task("dummy")
        self.timer = None

    def render(self):
        return self.p

    def update_progress(self):
        # TODO: this should pick actual progress of the song
        if self.p.finished:
            self.timer.stop_no_wait()
        else:
            self.p.advance(self.task_id, advance=self.INTERVAL)
        self.update()

    def pause(self):
        """Action on song pause to stop progress bar updates"""
        self.timer.pause()

    def resume(self):
        """Action on song play to resume progress bar updates"""
        self.timer.resume()

    def display_progress(self, song: Song):
        self.reset()

        self.task_id = self.p.add_task("play", total=int(song.duration))
        self.timer = self.set_interval(self.INTERVAL, self.update_progress)
        self.update()

    def reset(self):
        if self.task_id is not None:
            self.p.remove_task(self.task_id)

        # TODO: do we actually need this explicit tasks reset
        self.p._tasks = {}
        if self.timer is not None:
            self.timer.stop_no_wait()
            self._timers.discard(self.timer)



from typing import Optional

import click
from rich.console import Console

from pypod import config
from pypod.player import Pod
from pypod.app import PyPodApp

console = Console(highlight=False)


@click.group(invoke_without_command=True)
@click.version_option(message="pypod, version %(version)s")
@click.argument("filepath", metavar="path")
def cli(filepath):
    player = Pod()
    playlist = player.generate_playlist(filepath)
    player.load(playlist)
    app = PyPodApp(player=player)
    app.run()
    

if __name__ == "__main__":
    cli()


from threading import Thread, Event
from typing import Optional
from pathlib import Path

from pypod.song import Song, WAVSong


class Playlist:
    def __init__(self, name: str):
        self.name = name
        self.songs = []
        self._index = 0
        self._shifted = False

    def add_song(self, song):
        self.songs.append(song)

    @property
    def position(self):
        return self._index

    def __getitem__(self, key):
        return self.songs.__getitem__(key)

    def __setitem__(self, key, value):
        self.songs.__setitem__(key, value)

    def __iter__(self):
        return iter(self.songs)

    def __next__(self):
        if not self._shifted:
            self._shifted = True
            return self[0]
        index = (self._index + 1) % len(self)
        self._index = index
        return self.songs[self._index]

    def get_next(self):
        if not self._shifted:
            self._shifted = True
        return next(self)

    def get_prev(self):
        if not self._shifted:
            self._shifted = True
        index = (self._index - 1) % len(self)
        self._index = index
        return self[index]

    def __len__(self):
        return len(self.songs)


class Pod:
    def __init__(self):
        self._current: Song = None
        self._thread: Thread= None
        self.playlist: Playlist = None
        self.auto_finish: Event = Event()

    @staticmethod
    def generate_playlist(filepath: str | Path) -> Playlist:
        filepath = Path(filepath)
        playlist = Playlist("Default playlist")

        if filepath.is_dir():
            for song_path in sorted(filepath.glob("**/*.wav")):
                song = WAVSong(song_path.absolute())
                playlist.add_song(song)
        else:
            song = WAVSong(filepath.absolute())
            playlist.add_song(song)

        return playlist

    def load(self, playlist: Playlist):
        self.playlist = playlist

    def play_playlist(self):
        for song in self.playlist:
            t = Thread(target=song.play,)
            self._current = song
            t.start()
            break
        print("Finished whole playlist")

    def play(self):
        if self.is_playing:
            return

        print("Unpausing current song")
        # un-pause section
        song = self.song
        if song is not None and song.paused:
            song.paused = False
            return

        # play from playlist section
        song = next(self.playlist)
        self._play_song(song)

    def _play_wrap(self, song: Song):
        song.play()
        if not song.stopped:
            self.auto_finish.set()

    def _play_song(self, song):
        t = Thread(target=self._play_wrap, args=(song,))
        # t = Thread(target=song.play,)
        self._current = song
        print(f"Start playing {song} song")
        self.auto_finish.clear()
        t.start()
        self._thread = t

    def exit(self):
        if self.song is not None:
            print("Stopping current song")
            self.song.stop()
            self._thread.join()

    @property
    def is_playing(self) -> bool:
        return self._current is not None and not self._current.paused

    @property
    def song(self) -> Optional[Song]:
        """Return currently playing song"""
        return self._current

    def pause(self):
        if self.is_playing:
            print("Pausing current song")
            self.song.pause()

    def next(self):
        if self.song is not None:
            self.song.stop()
        song = self.playlist.get_next()
        self._play_song(song)

    def prev(self):
        if self.song is not None:
            self.song.stop()
        song = self.playlist.get_prev()
        self._play_song(song)

    def __del__(self):
        """Cleanup"""
        # TODO: save playlist/song state on exit?


def play_directory():
    from pypod import config
    player = Pod()
    playlist = player.generate_playlist(config.ASSETS_DIR)
    player.load(playlist)
    player.play()


if __name__ == "__main__":
    play_directory()



import time
import wave as pywave
import contextlib
from abc import ABC, abstractmethod
from pathlib import Path

import pyaudio


class Song(ABC):
    def __init__(self, filepath: Path):
        self.filepath: Path = filepath
        self.paused: bool = True
        self.stopped: bool = True

    @abstractmethod
    def play(self):
        pass

    def pause(self):
        self.paused = True

    def stop(self):
        self.stopped = True

    @property
    @abstractmethod
    def duration(self):
        pass

    def progress(self):
        pass

    @property
    def name(self):
        return self.filepath.stem

    def __str__(self):
        return self.name

    def __rich__(self):
        return f"♫ {self.name}"


class WAVSong(Song):

    CHUNK = 1024

    def __init__(self, filepath: Path):
        super().__init__(filepath=filepath)
        self.py_audio = pyaudio.PyAudio()

    def play(self):
        with pywave.open(str(self.filepath), "rb") as wave:
            format = self.py_audio.get_format_from_width(
                wave.getsampwidth()
            )
            stream = self.py_audio.open(
                format=format,
                channels=wave.getnchannels(),
                rate=wave.getframerate(),
                output=True,
            )
            self.paused = False
            self.stopped = False
            with contextlib.closing(stream):
                while True:
                    if self.stopped:
                        return
                    if self.paused:
                        if stream.is_active():
                            stream.stop_stream()
                        time.sleep(0.01)
                    else:
                        if stream.is_stopped():
                            stream.start_stream()
                        data = wave.readframes(self.CHUNK)
                        if not data:
                            break
                        stream.write(data)
                stream.stop_stream()
            self.paused = True

    def __del__(self):
        self.py_audio.terminate()

    @property
    def duration(self) -> float:
        with pywave.open(str(self.filepath), "rb") as wave:
            frames = wave.getnframes()
            rate = float(wave.getframerate())
        return frames / rate


if __name__ == "__main__":
    from pypod import config
    filepath = config.ASSETS_DIR / "rain_and_storm.wav"
    s1 = WAVSong(filepath=filepath)
    s1.play()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pypod']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'pyaudio>=0.2.12,<0.3.0',
 'rich>=12.6.0,<13.0.0',
 'textual>=0.6.0,<0.7.0']

entry_points = \
{'console_scripts': ['pypod = pypod.cli:cli']}

setup_kwargs = {
    'name': 'python-pod',
    'version': '0.1.0',
    'description': 'Python terminal music player',
    'long_description': '## PyPod\n\nPython console music player\n\n```bash\n$ brew install portaudio\n$ poetry install\n```\n\n\n### Usage\n```bash\n$ pypod <path-to-directory>\n$ pypod filename.wav\n```\n',
    'author': 'Misha Behersky',
    'author_email': 'bmwant@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
